#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "lib.h"
#include "formula.h"
#include "label.h"
#include "variable.h"
#include "command.h"

static struct command commands[COMMAND_MAXNUM];

static command_t pool = NULL;

struct command_ope opecodes[] = {
  { NULL      , NULL , COMMAND_NONE     },
  { NULL      , NULL , COMMAND_LABEL    },
  { "#"       , NULL , COMMAND_COMMENT  },
  { "RUN"     , "RN.", COMMAND_RUN      },
  { "EDIT"    , "ET.", COMMAND_EDIT     },
  { "END"     , NULL , COMMAND_END      },
  { "EXIT"    , NULL , COMMAND_EXIT     },
  { "QUIT"    , "QT.", COMMAND_EXIT     },
  { "BREAK"   , "BK.", COMMAND_BREAK    },
  { "STEP"    , "SE.", COMMAND_STEP     },
  { "STOP"    , "SP.", COMMAND_STOP     },
  { "SKIP"    , NULL , COMMAND_SKIP     },
  { "CONTINUE", "CN.", COMMAND_CONTINUE },
  { "LET"     , "LT.", COMMAND_LET      },
  { "PRINT"   , "PT.", COMMAND_PRINT    },
  { "GOTO"    , "GT.", COMMAND_GOTO     },
  { "GONEXT"  , "GN.", COMMAND_GONEXT   },
  { "GOSUB"   , "GS.", COMMAND_GOSUB    },
  { "RETURN"  , "RT.", COMMAND_RETURN   },
  { "LOOPEND" , "LE.", COMMAND_LOOPEND  },
  { "LOOP"    , "LP.", COMMAND_LOOP     },
  { "FOR"     , "FR.", COMMAND_FOR      },
  { "NEXT"    , "NX.", COMMAND_NEXT     },
  { "DIM"     , "DM.", COMMAND_DIM      },
  { "MEMORY"  , "MM.", COMMAND_MEMORY   },
  { "DISCARD" , "DC.", COMMAND_DISCARD  },
  { "NEW"     , "NW.", COMMAND_NEW      },
  { "LIST"    , "LS.", COMMAND_LIST     },
  { "SAVE"    , "SV.", COMMAND_SAVE     },
  { "LOAD"    , "LD.", COMMAND_LOAD     },
  { "WAIT"    , "WT.", COMMAND_WAIT     },
  { "DUMP"    , "DP.", COMMAND_DUMP     },
  { NULL      , NULL , COMMAND_LAST     }
};

int command_check(void)
{
  command_t command;
  int n = 0;

  for (command = pool; command && (n < COMMAND_MAXNUM); command = command->next)
    n++;

  if (command || (n != COMMAND_MAXNUM))
    return NLL_ERRCODE_COMMAND_NOT_EMPTY;

  return 0;
}

static int namecmp(const char *s0, const char *s1, int len)
{
  int i;
  for (i = 0; i < len; i++) {
    if (toupper(*(s0++)) != toupper(*(s1++)))
      return -1;
  }
  return 0;
}

int command_search(char *p, char **endp, command_ope_t *opep)
{
  command_ope_t ope;
  int i, len;

  for (ope = opecodes; ope->type != COMMAND_LAST; ope++) {
    if (ope->name) {
      len = strlen(ope->name);
      if (!namecmp(ope->name, p, len))
	goto found;
    }
  }

  for (ope = opecodes; ope->type != COMMAND_LAST; ope++) {
    if (ope->other_name) {
      len = strlen(ope->other_name);
      if (!namecmp(ope->other_name, p, len))
	goto found;
    }
  }

  for (ope = opecodes; ope->type != COMMAND_LAST; ope++) {
    if (ope->name) {
      len = strlen(ope->name);
      for (i = 1; i < len; i++) {
	if (!namecmp(ope->name, p, i) && (p[i] == '.')) {
	  len = i + 1;
	  goto found;
	}
      }
    }
  }

  return NLL_ERRCODE_COMMAND_NOT_FOUND;

found:
  *opep = ope;
  for (i = 0; i < len; i++, p++)
    *p = toupper(*p);
  if (endp)
    *endp = p;
  return 0;
}

int command_get(command_type_t type, command_ope_t *opep)
{
  command_ope_t ope;

  for (ope = opecodes; ope->type != COMMAND_LAST; ope++) {
    if (ope->type == type) {
      *opep = ope;
      return 0;
    }
  }

  return NLL_ERRCODE_COMMAND_NOT_FOUND;
}

int command_alloc(command_t *commandp)
{
  command_t command;

  if (!pool)
    return NLL_ERRCODE_COMMAND_BUFFER_OVER;

  command = pool;
  pool = pool->next;

  memset(command, 0, sizeof(*command));

  *commandp = command;

  return 0;
}

int command_free(command_t command)
{
  command_t next;

  for (; command; command = next) {
    next = command->next;

    command_clear(command);

    command->next = pool;
    pool = command;
  }

  return 0;
}

int command_clear(command_t command)
{
  command_arg_t arg;
  int i;

  if (command->prefix) {
    formula_free(&command->prefix);
    command->prefix = NULL;
  }

  if (command->ope) {
    switch (command->ope->type) {
    case COMMAND_LABEL:
      arg = &command->args[0];
      if (arg->arg.label.label) {
	label_clear(arg->arg.label.label);
	label_del(arg->arg.label.label);
      }
      if (arg->arg.label.args)
	formula_free(&arg->arg.label.args);
      break;

    default:
      for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
	arg = &command->args[i];
	switch (arg->type) {
	case COMMAND_ARG_LABEL:
	  if (arg->arg.label.label)
	    label_del(arg->arg.label.label);
	  if (arg->arg.label.args)
	    formula_free(&arg->arg.label.args);
	  break;
	case COMMAND_ARG_VARIABLE:
	  if (arg->arg.variable.variable)
	    variable_del(arg->arg.variable.variable);
	  break;
	case COMMAND_ARG_FORMULA:
	  if (arg->arg.formula.element)
	    formula_free(&arg->arg.formula.element);
	  break;
	case COMMAND_ARG_WORD:
	case COMMAND_ARG_UNKNOWN:
	default:
	  break;
	}
      }
      break;
    }
  }

  memset(command->args, 0, sizeof(command->args));

  command->ope = NULL;

  return 0;
}

int command_clean(command_t command)
{
  command_arg_t arg;
  int r, i;

  for (; command; command = command->next) {
    if (command->prefix) {
      if ((r = formula_clean(command->prefix)) < 0)
	return r;
    }

    if (command->ope) {
      switch (command->ope->type) {
      case COMMAND_LABEL:
	arg = &command->args[0];
	if (arg->arg.label.label) {
	  if ((r = label_clean(arg->arg.label.label)) < 0)
	    return r;
	}
	if (arg->arg.label.args) {
	  if ((r = formula_clean(arg->arg.label.args)) < 0)
	    return r;
	}
	break;

      default:
	for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
	  arg = &command->args[i];
	  switch (arg->type) {
	  case COMMAND_ARG_LABEL:
	    if (arg->arg.label.label) {
	      if ((r = label_clean(arg->arg.label.label)) < 0)
		return r;
	    }
	    if (arg->arg.label.args) {
	      if ((r = formula_clean(arg->arg.label.args)) < 0)
		return r;
	    }
	    break;
	  case COMMAND_ARG_VARIABLE:
	    if (arg->arg.variable.variable) {
	      if ((r = variable_clean(arg->arg.variable.variable)) < 0)
		return r;
	    }
	    break;
	  case COMMAND_ARG_FORMULA:
	    if (arg->arg.formula.element) {
	      if ((r = formula_clean(arg->arg.formula.element)) < 0)
		return r;
	    }
	    break;
	  case COMMAND_ARG_WORD:
	  case COMMAND_ARG_UNKNOWN:
	  default:
	    break;
	  }
	}
	break;
      }
    }
  }

  return 0;
}

int command_init(void)
{
  command_t command;
  int i;

  pool = NULL;
  memset(commands, 0, sizeof(commands));

  for (i = 0; i < COMMAND_MAXNUM; i++) {
    command = &commands[i];
    command->next = pool;
    pool = command;
  }

  return 0;
}

int command_dump(command_t command)
{
  int i;
  command_arg_t arg;
  char *s;

  for (; command; command = command->next) {
    printf("\t");
    if (command->errcode) {
      printf("ERROR\t");
      NLL_ERRPRINT(command->errcode, command->line);
      continue;
    }
    if (command->prefix) {
      formula_dump(command->prefix);
    }
    printf("\t");
    printf("%s", command->ope->name ? command->ope->name : "NONE");
    printf("\tArgs:");
    for (i = 0; i < COMMAND_ARG_MAXNUM; i++) {
      arg = &command->args[i];
      s = "";
      switch (arg->type) {
      case COMMAND_ARG_VARIABLE: if (arg->arg.variable.variable) s = " V"; break;
      case COMMAND_ARG_WORD:     if (arg->arg.word.word[0])      s = " W"; break;
      case COMMAND_ARG_LABEL:
	if (arg->arg.label.label) {
	  printf(" L:");
	  formula_dump(arg->arg.label.args);
	}
	break;

      case COMMAND_ARG_FORMULA:
	if (arg->arg.formula.element) {
	  printf(" F:");
	  formula_dump(arg->arg.formula.element);
	}
	break;

      case COMMAND_ARG_UNKNOWN:
      default:
	break;
      }
      printf("%s", s);
    }
    printf("\n");
  }

  return 0;
}
